import logging

from requests import Response

from django_swagger_tester.case.base import ResponseCaseTester, SchemaCaseTester
from django_swagger_tester.response_validation.base import ResponseTester
from django_swagger_tester.utils import unpack_response

logger = logging.getLogger('django_swagger_tester')


def response_validation(loader_class, response: Response, method: str, route: str, **kwargs) -> None:  # noqa: TYP001
    """
    Verifies that an OpenAPI schema definition matches an API response.

    It inspects the schema recursively, and verifies that the schema matches the structure of the response at every level.

    :param loader_class: Class containing a `get_response_schema` method
    :param response: HTTP response
    :param method: HTTP method ('get', 'put', 'post', ...)
    :param route: Relative path of the endpoint being tested
    :raises: django_swagger_tester.exceptions.SwaggerDocumentationError or django_swagger_tester.exceptions.CaseError
    """
    data, status_code = unpack_response(response)
    loader = loader_class(route=route, status_code=status_code, method=method)
    response_schema = loader.get_response_schema()
    ResponseTester(response_schema=response_schema, response_data=data, **kwargs)
    ResponseCaseTester(response_data=data, **kwargs)
    SchemaCaseTester(schema=response_schema, **kwargs)



import json
import logging

from django_swagger_tester.case.base import SchemaCaseTester
from django_swagger_tester.exceptions import SwaggerDocumentationError
from django_swagger_tester.input_validation.utils import get_request_body_schema, serialize_schema

logger = logging.getLogger('django_swagger_tester')


# noinspection PyUnboundLocalVariable
def input_validation(
    loader_class,  # noqa: TYP001
    serializer,  # noqa: TYP001
    method: str,
    route: str,
    camel_case_parser: bool,
    **kwargs,
) -> None:
    """
    Verifies that an OpenAPI schema request body definition is valid, according to the API view's input serializer.

    :param loader_class: Class containing a `get_request_body` method
    :param serializer: Serializer class used for input validation in your API view
    :param method: HTTP method ('get', 'put', 'post', ...)
    :param route: Relative path of the endpoint being tested
    :param camel_case_parser: True if request body should be camel cased - this is usually required when you're using
           djangorestframework-camel-case parses for your APIs.
    :raises: django_swagger_tester.exceptions.SwaggerDocumentationError or django_swagger_tester.exceptions.CaseError
    """
    loader = loader_class(route=route, method=method, **kwargs)
    endpoint_schema = loader.get_request_body()

    request_body_schema = get_request_body_schema(endpoint_schema)
    if 'example' in request_body_schema:
        # Find a ready dict object
        # This happens when you use your Serializer as the request body parameter in drf_yasg auto schemas
        example = request_body_schema['example']
    else:
        # Parses schema bit by bit
        # This happens if you use static schemas or document your request body using Schema objects in drf_yasg
        example = serialize_schema(request_body_schema)

    # Make camelCased input snake_cased so the serializer can read it
    if camel_case_parser:
        from djangorestframework_camel_case.util import underscoreize

        example = underscoreize(example)

    logger.debug('Validating example: %s', example)
    serializer = serializer(data=example)  # type: ignore
    if not serializer.is_valid():
        logger.info('Example was invalid')
        raise SwaggerDocumentationError(
            f'Request body is not valid according to the passed serializer.'
            f'\n\nSwagger example request body: \n\n\t{json.dumps(example)}'
            f'\n\nSerializer error:\n\n\t{json.dumps(serializer.errors)}\n\n'
            f'Note: If all your parameters are correct, you might need to change `camel_case_parser` to True or False.'
        )

    # Check the example's case for inconsistencies
    SchemaCaseTester(request_body_schema, **kwargs)



import logging

from django.conf import settings as django_settings
from django.core.exceptions import ImproperlyConfigured

logger = logging.getLogger('django_swagger_tester')


class SwaggerTesterSettings(object):
    """
    Loads and validates the packages Django settings.
    """

    def __init__(self) -> None:
        """
        Initializes tester class with base settings.
        """
        self.CASE = 'camel case'
        self.PATH = ''
        self.CAMEL_CASE_PARSER = False

        if not hasattr(django_settings, 'SWAGGER_TESTER'):
            return

        logger.debug('Loading settings.')
        _settings = django_settings.SWAGGER_TESTER

        for setting, value in _settings.items():
            if hasattr(self, setting):
                setattr(self, setting, value)
            else:
                logger.error('Found an excess key in the SWAGGER_TESTER settings: `%s`.', setting)
                raise ImproperlyConfigured(f'`{setting}` is not a valid setting for the django-swagger-tester module')

        logger.debug('Validating settings.')
        accepted_cases = [
            'camel case',
            'camelCase',
            'snake case',
            'snake_case',
            'kebab case',
            'kebab-case',
            'pascal case',
            'PascalCase',
        ]
        if self.CASE is None:
            pass  # <-- we skip case checks when CASE is None
        elif not isinstance(self.CASE, str) or self.CASE not in accepted_cases:
            logger.error('CASE setting is misspecified.')
            raise ImproperlyConfigured(
                f'The django-swagger-tester package currently doesn\'t support a case called {self.CASE}.'
                f' Set case to `snake case` for snake_case, '
                f'`camel case` for camelCase, '
                f'`pascal case` for PascalCase,'
                f'`kebab case` for kebab-case, '
                f'or to `None` to skip case validation outright.'
            )
        if not isinstance(self.CAMEL_CASE_PARSER, bool):
            raise ImproperlyConfigured(
                '`CAMEL_CASE_PARSER` needs to be True or False, or unspecified (defaults to False).'
            )
        else:
            if self.CAMEL_CASE_PARSER:
                try:
                    import djangorestframework_camel_case  # noqa: F401
                except ImportError:
                    raise ImproperlyConfigured(
                        'The package `djangorestframework_camel_case` is not installed, '
                        'and is required to enable camel case parsing.'
                    )


settings = SwaggerTesterSettings()



import json
import logging
import os
from typing import Optional

import yaml
from django.conf import settings
from django.core.exceptions import ImproperlyConfigured

from django_swagger_tester.input_validation.utils import get_request_body
from django_swagger_tester.response_validation.utils import get_response_schema
from django_swagger_tester.utils import resolve_path, validate_inputs

logger = logging.getLogger('django_swagger_tester')


# noinspection PyMethodMayBeStatic
class LoadStaticSchema:
    def __init__(self, route: str, method: str, status_code: Optional[int] = None, **kwargs) -> None:
        """
        Loads OpenAPI schema from a static file.

        :param route: a django-resolved endpoint path
        :param status_code: the relevant HTTP response status code to check in the OpenAPI schema
        :param method: the relevant HTTP method to check in the OpenAPI schema
        """
        validate_inputs(route=route, status_code=status_code, method=method)
        package_settings = settings.SWAGGER_TESTER
        self.validation(package_settings)
        self.path = package_settings['PATH']

        self.route = resolve_path(route)
        self.status_code = status_code
        self.method = method

    def validation(self, package_settings: dict) -> None:
        """
        Before trying to load static schema, we need to verify that:
        - The path to the static file is provided, and that the file type is compatible (json/yml/yaml)
        - The right parsing library is installed (pyYAML for yaml, json is builtin)
        """
        if 'PATH' not in package_settings or package_settings['PATH'] is None:
            logger.error('PATH setting is not specified.')
            raise ImproperlyConfigured(
                f'`PATH` is required when testing static schemas. Please update your SWAGGER_TESTER settings.'
            )
        elif not isinstance(package_settings['PATH'], str):
            logger.error('PATH setting is not a string.')
            raise ImproperlyConfigured('`PATH` needs to be a string. Please update your SWAGGER_TESTER settings.')
        if '.yml' in package_settings['PATH'] or '.yaml' in package_settings['PATH']:
            try:
                import yaml  # noqa: F401
            except ModuleNotFoundError:
                raise ImproperlyConfigured(
                    'The package `PyYAML` is required for parsing yaml files. '
                    'Please run `pip install PyYAML` to install it.'
                )

    def get_schema(self) -> dict:
        """
        Loads a static OpenAPI schema from file, and parses it to a python dict.

        :return: Schema contents as a dict
        :raises: ImproperlyConfigured
        """
        if not os.path.isfile(self.path):
            logger.error('Path `%s` does not resolve as a valid file.', self.path)
            raise ImproperlyConfigured(
                f'The path `{self.path}` does not point to a valid file. Make sure to point to the specification file.'
            )
        try:
            logger.debug('Fetching static schema from %s', self.path)
            with open(self.path, 'r') as f:
                content = f.read()
        except Exception as e:
            logger.exception('Exception raised when fetching OpenAPI schema from %s. Error: %s', self.path, e)
            raise ImproperlyConfigured(
                f'Unable to read the schema file. Please make sure the path setting is correct.\n\nError: {e}'
            )
        if '.json' in self.path:
            return json.loads(content)
        elif '.yaml' in self.path or '.yml' in self.path:
            return yaml.load(content, Loader=yaml.FullLoader)
        else:
            raise ImproperlyConfigured('The specified file path does not seem to point to a JSON or YAML file.')

    def get_response_schema(self) -> dict:
        """
        Indexes schema by url, HTTP method, and status code to get the section of a schema related to a specific response.
        """
        schema = self.get_schema()
        return get_response_schema(schema=schema, method=self.method, status_code=self.status_code, route=self.route)

    def get_request_body(self) -> dict:
        """
        Indexes schema by url, HTTP method, and status code,
        to get the request body of the section of a schema related to an endpoint.
        """
        schema = self.get_schema()
        return get_request_body(schema=schema, method=self.method, route=self.route)



import logging
import re
from typing import Any, Callable, Union

from django_swagger_tester.exceptions import CaseError

logger = logging.getLogger('django_swagger_tester')


def case_check(case: Union[str, None]) -> Callable[[str, str], None]:
    """
    Returns the appropriate case check based on the `case` input parameter.

    Validation for accepted `case` inputs should be done in the package configuration.

    :param case: str
    :return: callable function
    """
    logger.debug('Returning `%s` case function', case)
    return {  # type: ignore
        'camel case': is_camel_case,
        'camelCase': is_camel_case,
        'snake case': is_snake_case,
        'snake_case': is_snake_case,
        'kebab case': is_kebab_case,
        'kebab-case': is_kebab_case,
        'pascal case': is_pascal_case,
        'PascalCase': is_pascal_case,
        None: skip,
    }[case]


def is_camel_case(key: str) -> None:
    """
    Asserts that a value is camelCased.

    :param key: str
    :return: None
    :raises: django_swagger_tester.exceptions.CaseError
    """
    logger.debug('Verifying that `%s` is properly camel cased', key)
    if len(key) == 0:
        return
    if len(key) == 1 and (key.isalpha() is False or (key.isalpha() is True and key != key.casefold())):
        logger.error('`%s` is not properly camel cased', key)
        raise CaseError(f'The property `{key}` is not properly camelCased')
    else:
        camel_cased_key = key[0].lower() + re.sub(r'[\-_.\s]([a-z])', lambda matched: matched.group(1).upper(), key[1:])
        if camel_cased_key != key:
            logger.error('`%s` is not properly camel cased', key)
            raise CaseError(f'The property `{key}` is not properly camelCased')


def is_snake_case(key: str) -> None:
    """
    Asserts that a value is snake_cased.

    :param key: str
    :return: None
    :raises: django_swagger_tester.exceptions.CaseError
    """
    logger.debug('Verifying that `%s` is properly snake cased', key)
    if len(key) == 0:
        return
    if len(key) == 1 and (key.isalpha() is False or (key.isalpha() is True and key != key.casefold())):
        logger.error('%s is not snake cased', key)
        raise CaseError(f'The property `{key}` is not properly snake_cased')
    snake_cased_key = (
        re.sub('([a-z0-9])([A-Z])', r'\1_\2', re.sub('(.)(-)([A-Z][a-z]+)', r'\1_\2', key))
        .lower()
        .replace('-', '_')
        .replace(' ', '')
    )
    if snake_cased_key != key:
        logger.error('%s is not snake cased', key)
        raise CaseError(f'The property `{key}` is not properly snake_cased')


def is_kebab_case(key: str) -> None:
    """
    Asserts that a value is kebab-cased.

    :param key: str
    :return: None
    :raises: django_swagger_tester.exceptions.CaseError
    """
    logger.debug('Verifying that `%s` is properly kebab cased', key)
    if len(key) == 0:
        return
    if len(key) == 1 and (key.isalpha() is False or (key.isalpha() is True and key != key.casefold())):
        logger.error('%s is not kebab cased', key)
        raise CaseError(f'The property `{key}` is not properly kebab-cased')
    kebab_cased_key = (
        re.sub('([a-z0-9])([A-Z])', r'\1-\2', re.sub('(.)([A-Z][a-z]+)', r'\1-\2', key))
        .lower()
        .replace('_', '-')
        .replace(' ', '')
    )
    if kebab_cased_key != key:
        logger.error('%s is not kebab cased', key)
        raise CaseError(f'The property `{key}` is not properly kebab-cased')


def is_pascal_case(key: str) -> None:
    """
    Asserts that a value is PascalCased.

    :param key: str
    :return: None
    :raises: django_swagger_tester.exceptions.CaseError
    """
    logger.debug('Verifying that `%s` is properly pascal cased', key)
    if len(key) == 0:
        return
    if len(key) == 1 and (key.isalpha() is False or (key.isalpha() is True and key != key.upper())):
        logger.error('%s is not pascal cased', key)
        raise CaseError(f'The property `{key}` is not properly PascalCased')
    pascal_cased_key = key[0].upper() + re.sub(r'[\-_.\s]([a-z])', lambda matched: matched.group(1).upper(), key[1:])
    if pascal_cased_key != key:
        logger.error('%s is not pascal cased', key)
        raise CaseError(f'The property `{key}` is not properly PascalCased')


def skip(*args: Any) -> None:
    """
    Skips case assertion.

    :return: None
    """
    logger.debug('Skipping case validation')
    pass



import difflib
import logging
from typing import List, Tuple, Optional

from django.core.exceptions import ImproperlyConfigured
from django.urls import Resolver404, resolve
from requests import Response
from rest_framework.schemas.generators import EndpointEnumerator

logger = logging.getLogger('django_swagger_tester')


def get_paths() -> List[str]:
    """
    Returns a list of endpoint paths.
    """
    return list(set(endpoint[0] for endpoint in EndpointEnumerator().get_api_endpoints()))  # noqa: C401


def resolve_path(endpoint_path: str) -> str:
    """
    Resolves a Django path.
    """
    try:
        logger.debug('Resolving path.')
        if endpoint_path == '' or endpoint_path[0] != '/':
            logger.debug('Adding leading `/` to provided path')
            endpoint_path = '/' + endpoint_path
        try:
            resolved_route = resolve(endpoint_path)
            logger.debug('Resolved %s successfully', endpoint_path)
        except Resolver404:
            resolved_route = resolve(endpoint_path + '/')
            endpoint_path = (
                endpoint_path + '/'
            )  # if we don't change endpoint path here, indexing paths will fail later on
            logger.warning('Endpoint path is missing a trailing slash: %s', endpoint_path)

        kwarg = resolved_route.kwargs
        for key, value in kwarg.items():
            # Replacing kwarg values back into the string seems to be the simplest way of bypassing complex regex handling
            # However, its important not to freely use the .replace() function, as a {value} of `1` would also cause the `1` in api/v1/ to
            # be replaced
            var_index = endpoint_path.rfind(value)
            endpoint_path = endpoint_path[:var_index] + f'{{{key}}}' + endpoint_path[var_index + len(value) :]
        return endpoint_path

    except Resolver404:
        logger.error(f'URL `%s` did not resolve successfully', endpoint_path)
        paths = get_paths()
        closest_matches = ''.join([f'\n- {i}' for i in difflib.get_close_matches(endpoint_path, paths)])
        if closest_matches:
            raise ValueError(
                f'Could not resolve path `{endpoint_path}`.\n\nDid you mean one of these?{closest_matches}\n\n'
                f'If your path contains path parameters (e.g., `/api/<version>/...`), make sure to pass a '
                f'value, and not the parameter pattern.'
            )
        raise ValueError(f'Could not resolve path `{endpoint_path}`')


def validate_method(method: str) -> str:
    """
    Validates a string as a HTTP method.
    """
    methods = ['get', 'post', 'put', 'patch', 'delete', 'options', 'head']
    if not isinstance(method, str) or method.lower() not in methods:
        logger.error('Method `%s` is invalid. Should be one of: %s.', method, ', '.join([i.upper() for i in methods]))
        raise ImproperlyConfigured(
            f'Method `{method}` is invalid. Should be one of: {", ".join([i.upper() for i in methods])}.'
        )
    return method


def unpack_response(response: Response) -> Tuple[dict, int]:
    """
    Unpacks HTTP response.
    """
    try:
        status_code = response.status_code
    except Exception as e:
        logger.exception('Unable to open response object')
        raise ValueError(
            f'Unable to unpack response object. Make sure you are passing response, and not response.json(). Error: {e}'
        )
    if hasattr(response, 'json'):
        return response.json(), status_code
    else:
        raise ImproperlyConfigured(
            'Response does not contain a JSON-formatted response and cannot be tested against a response schema.'
        )


def replace_refs(schema: dict) -> dict:
    """
    Finds all $ref sections in a schema and replaces them with the referenced content.
    This way we only have to worry about $refs once.

    :param schema: OpenAPI schema
    :return Adjusted OpenAPI schema
    """
    if '$ref' not in str(schema):
        return schema

    def find_and_replace_refs_recursively(d: dict, schema: dict) -> dict:
        """
        Iterates over a dictionary to look for pesky $refs.
        """
        if '$ref' in d:
            indices = [i for i in d['$ref'][d['$ref'].index('#') + 1 :].split('/') if i]
            temp_schema = schema
            for index in indices:
                logger.debug(f'indexing by %s', index)
                temp_schema = temp_schema[index]
            return temp_schema
        for k, v in d.items():
            if isinstance(v, list):
                d[k] = iterate_list(v, schema)
            elif isinstance(v, dict):
                d[k] = find_and_replace_refs_recursively(v, schema)
        return d

    def iterate_list(l: list, s: dict) -> list:
        """
        Loves to iterate lists.
        """
        x = []
        for i in l:
            if isinstance(i, list):
                x.append(iterate_list(i, s))
            elif isinstance(i, dict):
                x.append(find_and_replace_refs_recursively(i, s))  # type: ignore
            else:
                x.append(i)
        return x

    return find_and_replace_refs_recursively(schema, schema)


def validate_inputs(route: str, status_code: Optional[int], method: str) -> None:
    """
    Input validation for response validator classes.

    :param route: a django-resolved endpoint path
    :param status_code: the relevant HTTP response status code to check in the OpenAPI schema
    :param method: the relevant HTTP method to check in the OpenAPI schema
    :raises: ImproperlyConfigured
    """
    if not isinstance(route, str):
        raise ImproperlyConfigured('`route` is invalid.')
    validate_method(method)
    if status_code is not None:
        if not isinstance(status_code, int):
            raise ImproperlyConfigured('`status_code` should be an integer.')
        if not 100 <= status_code <= 505:
            raise ImproperlyConfigured('`status_code` should be a valid HTTP response code.')



import logging
from typing import Callable, List

logger = logging.getLogger('django_swagger_tester')


def set_ignored_keys(**kwargs) -> List[str]:
    """
    Lets users pass a list of string that will not be checked by case-check.
    For example, validate_response(..., ignore_case=["List", "OF", "improperly cased", "kEYS"]).
    """
    if 'ignore_case' in kwargs:
        return kwargs['ignore_case']
    return []


def conditional_check(key: str, function: Callable, ignored_keys: list) -> None:
    """
    Checks a keys case if the key is not ignored.

    Put this in its own function so that we're consistent in response and schema validation handling.

    :param key: dictionary key
    :param function: case check callable
    :param ignored_keys: list of ignored values - values that shouldn't be checked
    raises: django_swagger_tester.exceptions.CaseError
    """
    if key not in ignored_keys:
        function(key)
    else:
        logger.debug('Skipping case check for ignored key `%s`', key)



import logging
from typing import Any

from django_swagger_tester.case.checks import case_check
from django_swagger_tester.case.utils import conditional_check, set_ignored_keys
from django_swagger_tester.configuration import settings
from django_swagger_tester.openapi import read_items, read_properties, read_type

logger = logging.getLogger('django_swagger_tester')


class ResponseCaseTester(object):
    """
    Iterates through an API response objects to verify that dict keys are cased correctly.
    The case we're checking for depends on the projects SWAGGER_TESTER `CASE` setting.
    """

    def __init__(self, response_data: Any, **kwargs) -> None:
        """
        Finds the appropriate case check function and calls the appropriate function base on the response datas type.

        :param response_data: typically will be an API responses response.json() output.
        """
        self.case_check = case_check(settings.CASE)
        self.ignored_keys = set_ignored_keys(**kwargs)
        if isinstance(response_data, dict):
            self.test_dict(response_data)
        elif isinstance(response_data, list):
            self.test_list(response_data)
        else:
            logger.debug('Skipping case check')

    def test_dict(self, dictionary: dict) -> None:
        """
        Iterates through a response dictionary to check keys' case, and to pass nested values for further test_checks.
        """
        if not isinstance(dictionary, dict):
            raise ValueError(f'Expected dictionary, but received {type(dictionary)}')
        for key, value in dictionary.items():
            conditional_check(key, self.case_check, self.ignored_keys)
            if isinstance(value, dict):
                self.test_dict(dictionary=value)
            elif isinstance(value, list):
                self.test_list(items=value)

    def test_list(self, items: list) -> None:
        """
        Iterates through a response list to pass appropriate nested items for further test_checks.
        Only dictionary keys need case checking, so that's what we're looking for.

        :param items: list of unknown items
        """
        if not isinstance(items, list):
            raise ValueError(f'Expected list, but received {type(items)}')
        for item in items:
            if isinstance(item, dict):
                self.test_dict(dictionary=item)
            elif isinstance(item, list):
                self.test_list(items=item)


class SchemaCaseTester(object):
    """
    Iterates through an OpenAPI schema to verify that object keys are cased correctly.
    The case we're checking for depends on the projects SWAGGER_TESTER `CASE` setting.
    """

    def __init__(self, schema: dict, **kwargs) -> None:
        """
        Finds the appropriate case check function and calls the appropriate function base on the schema item type.

        :param schema: openapi schema item
        """
        self.case_check = case_check(settings.CASE)
        self.ignored_keys = set_ignored_keys(**kwargs)
        if read_type(schema) == 'object':
            logger.debug('root -> dict')
            self.test_dict(schema)
        elif read_type(schema) == 'array':
            logger.debug('root -> list')
            self.test_list(schema)
        else:
            logger.debug('Skipping case check')

    def test_dict(self, obj: dict) -> None:
        """
        Iterates through a schema object to check keys' case, and to pass nested values for further test_checks.
        """
        properties = read_properties(obj)
        for key, value in properties.items():
            conditional_check(key, self.case_check, self.ignored_keys)
            if read_type(value) == 'object':
                logger.debug('dict -> dict')
                self.test_dict(obj=value)
            elif read_type(value) == 'array':
                logger.debug('dict -> list')
                self.test_list(array=value)

    def test_list(self, array: dict) -> None:
        """
        Iterates through a schema array to pass appropriate nested items for further test_checks.
        Only object keys need case checking, so that's what we're looking for.
        """
        item = read_items(array)
        if read_type(item) == 'object':
            logger.debug('list -> dict')
            self.test_dict(obj=item)
        elif read_type(item) == 'array':
            logger.debug('list -> list')
            self.test_list(array=item)



"""
This file should contain a collection of schema-parsing utility functions.

While this package should never become an OpenAPI schema parser,
it is useful for us to apply the rules of the schema specification
in case we're ever dealt with an incorrect schema.

Instead of raising unhandled errors, it is useful for us to raise appropriate exceptions.
"""
import logging
from typing import List

from django_swagger_tester.exceptions import OpenAPISchemaError, SwaggerDocumentationError

logger = logging.getLogger('django_swagger_tester')


def read_items(array: dict) -> dict:
    """
    Accesses the `items` attribute of an array.

    Rule: `items – must be present if type is array. The item schema must be an OpenAPI schema and not a standard JSON`
    https://swagger.io/docs/specification/data-models/keywords/

    :param array: schema array
    :return: array items
    :raises: django_swagger_tester.exceptions.OpenAPISchemaError
    """
    if 'items' not in array:
        raise OpenAPISchemaError(f'Array is missing an `items` attribute.\n\nArray schema: {array}')
    return array['items']


def list_types() -> List[str]:
    """
    Returns supported item types.
    """
    return ['string', 'boolean', 'integer', 'number', 'file', 'object', 'array']


def read_type(item: dict) -> str:
    """
    Accesses the `items` attribute of a schema item.

    Rule: `type – the value must be a single type and not an array of types.
            null is not supported as a type, use the`nullable: true keyword instead.`

    :param item: schema item
    :return: schema item type
    :raises: django_swagger_tester.exceptions.OpenAPISchemaError
    """
    if (
        item is None
        or not isinstance(item, dict)
        or 'type' not in item
        or not item['type']
        or not isinstance(item['type'], str)
    ):
        raise OpenAPISchemaError(
            f'Schema item has an invalid `type` attribute. The type should be a single string.\n\nSchema item: {item}'
        )
    if not item['type'] in list_types():
        raise OpenAPISchemaError(
            f'Schema item has an invalid `type` attribute. '
            f'The type `{item["type"]}` is not supported.\n\nSchema item: {item}'
        )
    return item['type']


def read_additional_properties(schema_object: dict) -> dict:
    """
    Accesses the `additionalProperties` attribute of a schema object.

    :param schema_object: schema object (dict)
    :return: schema object additional properties
    :raises: django_swagger_tester.exceptions.OpenAPISchemaError
    """
    if 'additionalProperties' not in schema_object:
        raise OpenAPISchemaError(
            f'Object is missing a `additionalProperties` attribute.\n\nObject schema: {schema_object}'
        )
    return schema_object['additionalProperties']


def read_properties(schema_object: dict) -> dict:
    """
    Accesses the `properties` attribute of a schema object.

    :param schema_object: schema object (dict)
    :return: schema object properties
    :raises: django_swagger_tester.exceptions.OpenAPISchemaError
    """
    if 'properties' not in schema_object:
        if 'additionalProperties' in schema_object:
            # We return this with an empty key, so we can still iterate over the results .items(), as we would with
            # normal properties
            return {'': read_additional_properties(schema_object)}
        raise OpenAPISchemaError(f'Object is missing a `properties` attribute.\n\nObject schema: {schema_object}')
    return schema_object['properties']


def is_nullable(schema_item: dict) -> bool:
    """
    Checks if the item is nullable.

    OpenAPI does not have a null type, like a JSON schema,
    but in OpenAPI 3.0 they added `nullable: true` to specify that the value may be null.
    Note that null is different from an empty string "".

    This feature was back-ported to the Swagger 2 parser as a vendored extension `x-nullable`. This is what drf_yasg generates.

    OpenAPI 3 ref: https://swagger.io/docs/specification/data-models/data-types/#null
    Swagger 2 ref: https://help.apiary.io/api_101/swagger-extensions/

    :param schema_item: schema item
    :return: whether or not the item can be None
    """
    openapi_schema_3_nullable = 'nullable'
    swagger_2_nullable = 'x-nullable'
    for nullable_key in [openapi_schema_3_nullable, swagger_2_nullable]:
        if schema_item and isinstance(schema_item, dict):
            if nullable_key in schema_item:
                if isinstance(schema_item[nullable_key], str):
                    if schema_item[nullable_key] == 'true':
                        return True
    return False


def index_schema(schema: dict, variable: str, error_addon: str = None) -> dict:
    """
    Indexes schema by string variable.

    :param schema: Schema to index
    :param variable: Variable to index by
    :param error_addon: Additional error info
    :return: Indexed schema
    :raises: django_swagger_tester.exceptions.SwaggerDocumentationError
    """
    if error_addon is None:
        error_addon = ''
    try:
        logger.debug('Indexing schema by `%s`', variable)
        return schema[f'{variable}']
    except KeyError:
        raise SwaggerDocumentationError(
            f'Failed indexing schema.\n\nError: Unsuccessfully tried to index the OpenAPI schema by `{variable}`.'
            + error_addon
        )



import logging
from typing import Union

from django.core.exceptions import ImproperlyConfigured

from django_swagger_tester.exceptions import OpenAPISchemaError
from django_swagger_tester.openapi import index_schema, read_items, read_type
from django_swagger_tester.utils import replace_refs

logger = logging.getLogger('django_swagger_tester')


def get_request_body(schema: dict, method: str, route: str) -> dict:
    """
    Indexes schema to get an endpoints request body.

    :param schema: Full OpenAPI schema
    :param method: HTTP request method
    :param route: Schema-compatible path
    :return: Request body schema
    """
    no_ref_schema = replace_refs(schema)
    paths_schema = index_schema(schema=no_ref_schema, variable='paths')
    route_error = f'\n\nFor debugging purposes: valid routes include {", ".join([key for key in paths_schema.keys()])}'
    route_schema = index_schema(schema=paths_schema, variable=route, error_addon=route_error)
    joined_methods = ', '.join([method.upper() for method in route_schema.keys() if method.upper() != 'PARAMETERS'])
    method_error = f'\n\nAvailable methods include {joined_methods}.'
    method_schema = index_schema(schema=route_schema, variable=method.lower(), error_addon=method_error)
    return index_schema(schema=method_schema, variable='parameters')


def get_request_body_schema(request_body_schema: dict) -> dict:
    """
    Translates the `parameters` section of an OpenAPI schema to an example request body.

    :param request_body_schema: Schema section
    :return: Example request body
    """
    try:
        schema = request_body_schema[0]
    except IndexError:
        raise OpenAPISchemaError(
            f'Request body does not seem to be documented. Schema parameters: {request_body_schema}'
        )
    if 'in' not in schema or schema['in'] != 'body':
        logger.debug('Request body schema seems to be missing a request body section')
        raise ImproperlyConfigured(
            'Tried to test request body documentation, but the provided schema has no request body.'
        )
    return schema['schema']


def _iterate_schema_dict(d: dict) -> dict:
    x = {}
    for key, value in d['properties'].items():
        if read_type(value) == 'object':
            x[key] = _iterate_schema_dict(value)
        elif read_type(value) == 'array':
            x[key] = _iterate_schema_list(value)  # type: ignore
        elif 'example' in value:
            x[key] = value['example']
        else:
            raise ImproperlyConfigured(f'This schema item does not seem to have example value. Item: {value}')
    return x


def _iterate_schema_list(l: dict) -> list:
    x = []
    i = read_items(l)
    if read_type(i) == 'object':
        x.append(_iterate_schema_dict(i))
    elif read_type(i) == 'array':
        x.append(_iterate_schema_list(i))  # type: ignore
    elif 'example' in i:
        x.append(i['example'])
    else:
        raise ImproperlyConfigured(f'This schema item does not seem to have example value. Item: {i}')
    return x


def serialize_schema(schema: dict) -> Union[list, dict, str, int, bool]:
    """
    Converts an OpenAPI schema representation of a dict to dict.
    """
    if read_type(schema) == 'array':
        logger.debug('--> list')
        return _iterate_schema_list(schema)
    elif read_type(schema) == 'object':
        logger.debug('--> dict')
        return _iterate_schema_dict(schema)
    elif 'example' in schema:
        return schema['example']
    else:
        raise ImproperlyConfigured(f'This schema item does not seem to have example value. Item: {schema}')



import logging
from django.core.exceptions import ImproperlyConfigured
from typing import Any, Union

from django_swagger_tester.openapi import is_nullable, list_types, read_items, read_properties, read_type
from django_swagger_tester.response_validation.utils import check_keys_match, format_error

logger = logging.getLogger('django_swagger_tester')


# noinspection PyMethodMayBeStatic
class ResponseTester:
    def __init__(self, response_schema: dict, response_data: Any, **kwargs) -> None:
        """
        Iterates through both a response schema and an actual API response to check that they match.

        :param response_schema: Response OpenAPI schema section
        :param response_data: API response data
        raises: django_swagger_tester.exceptions.SwaggerDocumentationError or ImproperlyConfigured
        """
        if '$ref' in str(response_schema):
            # `$ref`s should be replaced inplace before the schema is passed to this class
            raise ImproperlyConfigured(
                'Received invalid schema. You must replace $ref sections before passing a schema for validation.'
            )

        if read_type(response_schema) == 'object':
            logger.debug('init --> dict')
            self.test_dict(schema=response_schema, data=response_data, reference='init', **kwargs)
        elif read_type(response_schema) == 'array':
            logger.debug('init --> list')
            self.test_list(schema=response_schema, data=response_data, reference='init', **kwargs)
        # this should always be third, as list_types also contains `array` and `object`
        elif read_type(response_schema) in list_types():
            logger.debug('init --> item')
            self.test_item(schema=response_schema, data=response_data, reference='init', **kwargs)

    def test_dict(self, schema: dict, data: Union[list, dict], reference: str, **kwargs) -> None:
        """
        Verifies that a schema dict matches a response dict.

        :param schema: OpenAPI schema
        :param data: Response data
        :param reference: string reference pointing to function caller
        :raises: django_swagger_tester.exceptions.SwaggerDocumentationError
        """
        if not isinstance(data, dict):
            hint = ''
            if isinstance(data, list):
                hint = 'The expected item should be a dict, or your schema should be a list.'
            elif data is None:
                if 'x-nullable' in schema and schema['x-nullable']:
                    # NoneTypes are OK if the schema says the field is nullable
                    return
                hint = (
                    'If you wish to allow null values for this schema item, your schema needs to set `x-nullable: True`.'
                    '\nFor drf-yasg implementations, set `x_nullable=True` in your Schema definition.'
                )
            raise format_error(
                error_message=f"Mismatched types. Expected response to be <class 'dict'> but found {type(data)}.",
                data=data,
                schema=schema,
                reference=reference,
                hint=hint,
                **kwargs,
            )

        properties = read_properties(schema)
        schema_keys = properties.keys()
        response_keys = data.keys()
        check_keys_match(schema_keys, response_keys, schema, data, reference)

        for schema_key, response_key in zip(schema_keys, response_keys):

            # Check that each element in the schema exists in the response, and vice versa
            if schema_key not in response_keys:
                raise format_error(
                    error_message=f'Schema key `{schema_key}` was not found in the API response.',
                    data=data,
                    schema=schema,
                    reference=reference,
                    hint='You need to add the missing schema key to the response, or remove it from the documented response.',
                    **kwargs,
                )
            elif response_key not in schema_keys:
                raise format_error(
                    error_message=f'Response key `{response_key}` not found in the OpenAPI schema.',
                    data=data,
                    schema=schema,
                    reference=reference,
                    hint='You need to add the missing schema key to your documented response, or stop returning it in your API.',
                    **kwargs,
                )

            # Pass nested elements to the appropriate function
            schema_value = properties[schema_key]
            response_value = data[schema_key]

            if read_type(schema_value) == 'object':
                logger.debug('test_dict --> test_dict. Response: %s, Schema: %s', response_value, schema_value)
                self.test_dict(
                    schema=schema_value, data=response_value, reference=f'{reference}.dict:key:{schema_key}', **kwargs
                )
            elif read_type(schema_value) == 'array':
                logger.debug('test_dict --> test_list. Response: %s, Schema: %s', response_value, schema_value)
                self.test_list(
                    schema=schema_value, data=response_value, reference=f'{reference}.dict:key:{schema_key}', **kwargs
                )
            elif read_type(schema_value) in list_types():  # This needs to come after array and object test_checks
                logger.debug('test_dict --> test_item. Response: %s, Schema: %s', response_value, schema_value)
                self.test_item(
                    schema=schema_value, data=response_value, reference=f'{reference}.dict:key:{schema_key}', **kwargs
                )

    def test_list(self, schema: dict, data: Union[list, dict], reference: str, **kwargs) -> None:
        """
        Verifies that a schema array matches a response list.

        :param schema: OpenAPI schema
        :param data: Response data
        :param reference: string reference pointing to function caller
        :raises: django_swagger_tester.exceptions.SwaggerDocumentationError
        """
        if not isinstance(data, list):
            hint = ''
            if isinstance(data, dict):
                hint = 'You might need to wrap your response item in a list, or remove the excess list layer from your documented response.'
            elif data is None:
                if 'x-nullable' in schema and schema['x-nullable']:
                    # NoneTypes are OK if the schema says the field is nullable
                    return
            hint = (
                'If you wish to allow null values for this schema item, your schema needs to set `x-nullable: True`.'
                '\nFor drf-yasg implementations, set `x_nullable=True` in your Schema definition.'
            )
            raise format_error(
                error_message=f"Mismatched types. Expected response to be <class 'list'> but found {type(data)}.",
                data=data,
                schema=schema,
                reference=reference,
                hint=hint,
                **kwargs,
            )

        item = read_items(schema)
        for index in range(len(data)):

            if read_type(item) == 'object':
                logger.debug('test_list --> test_dict')
                self.test_dict(schema=item, data=data[index], reference=f'{reference}.list', **kwargs)

            elif read_type(item) == 'array':
                logger.debug('test_list --> test_dict')
                self.test_list(schema=item, data=data[index], reference=f'{reference}.list', **kwargs)

            elif read_type(item) in list_types():
                logger.debug('test_list --> test_item')
                self.test_item(schema=item, data=data[index], reference=f'{reference}.list', **kwargs)

    def test_item(self, schema: dict, data: Any, reference: str, **kwargs) -> None:
        """
        Verifies that a response value matches the example value in the schema.

        :param schema: OpenAPI schema
        :param data: response data item
        :param reference: string reference pointing to function caller
        :raises: django_swagger_tester.exceptions.SwaggerDocumentationError
        """
        checks = {
            'boolean': {
                'check': not isinstance(data, bool)
                and not (isinstance(data, str) and (data.lower() == 'true' or data.lower() == 'false')),
                'type': "<class 'bool'>",
            },
            'string': {'check': not isinstance(data, str) and data is not None, 'type': "<class 'str'>"},
            'integer': {'check': not isinstance(data, int) and data is not None, 'type': "<class 'int'>"},
            'number': {'check': not isinstance(data, float) and data is not None, 'type': "<class 'float'>"},
            'file': {'check': not isinstance(data, str) and data is not None, 'type': "<class 'str'>"},
        }
        if data is None and is_nullable(schema):
            return
        elif checks[schema['type']]['check']:
            raise format_error(
                error_message=f'Mismatched types.',
                data=data,
                schema=schema,
                reference=reference,
                hint=f'You need to change the response value type to {checks[schema["type"]]["type"]}, '
                f'or change your documented type to {type(data)}.',
                **kwargs,
            )



from rest_framework.response import Response

from django_swagger_tester.configuration import settings
from django_swagger_tester.input_validation.validation import input_validation as _input_validation
from django_swagger_tester.response_validation.validation import response_validation as _response_validation
from django_swagger_tester.static_schema.loader import LoadStaticSchema as _loader_class


def validate_response(response: Response, method: str, route: str, **kwargs) -> None:
    """
    Calls the response validation function with the static schema loader class.

    :param response: HTTP response
    :param method: HTTP method ('get', 'put', 'post', ...)
    :param route: Relative path of the endpoint being tested
    :raises: django_swagger_tester.exceptions.SwaggerDocumentationError or django_swagger_tester.exceptions.CaseError
    """
    return _response_validation(loader_class=_loader_class, response=response, method=method, route=route, **kwargs)


def validate_input(
    serializer, method: str, route: str, camel_case_parser: bool = settings.CAMEL_CASE_PARSER, **kwargs  # noqa: TYP001
) -> None:
    """
    Calls the input validation function with the static schema loader class.

    :param serializer: Serializer class used for input validation in your API view
    :param method: HTTP method ('get', 'put', 'post', ...)
    :param route: Relative path of the endpoint being tested
    :param camel_case_parser: True if request body should be camel cased - this is usually required when you're using
           djangorestframework-camel-case parses for your APIs.
    :raises: django_swagger_tester.exceptions.SwaggerDocumentationError or django_swagger_tester.exceptions.CaseError
    """
    return _input_validation(
        loader_class=_loader_class,
        serializer=serializer,
        method=method,
        route=route,
        camel_case_parser=camel_case_parser,
        **kwargs,
    )



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['django_swagger_tester',
 'django_swagger_tester.case',
 'django_swagger_tester.drf_yasg',
 'django_swagger_tester.input_validation',
 'django_swagger_tester.response_validation',
 'django_swagger_tester.static_schema']

package_data = \
{'': ['*']}

install_requires = \
['django>=2.2,<4.0', 'djangorestframework']

setup_kwargs = {
    'name': 'django-swagger-tester',
    'version': '1.0.3',
    'description': 'Django test utility for validating Swagger documentation',
    'long_description': 'Django Swagger Tester\n=====================\n\n.. role:: python(code)\n   :language: python\n\n.. image:: https://img.shields.io/pypi/v/django-swagger-tester.svg\n    :target: https://pypi.org/project/django-swagger-tester/\n\n.. image:: https://img.shields.io/pypi/pyversions/django-swagger-tester.svg\n    :target: https://pypi.org/project/django-swagger-tester/\n\n.. image:: https://img.shields.io/pypi/djversions/django-swagger-tester.svg\n    :target: https://pypi.python.org/pypi/django-swagger-tester\n\n.. image:: https://readthedocs.org/projects/django-swagger-tester/badge/?version=latest\n    :target: https://django-swagger-tester.readthedocs.io/en/latest/?badge=latest\n    :alt: Documentation Status\n\n.. image:: https://codecov.io/gh/sondrelg/django-swagger-tester/branch/master/graph/badge.svg\n    :target: https://codecov.io/gh/sondrelg/django-swagger-tester\n\n|\n\n.. image:: https://img.shields.io/badge/code%20style-black-000000.svg\n    :target: https://pypi.org/project/django-swagger-tester/\n\n.. image:: http://www.mypy-lang.org/static/mypy_badge.svg\n    :target: http://mypy-lang.org/\n    :alt: Checked with mypy\n\n.. image:: https://img.shields.io/badge/pre--commit-enabled-brightgreen?logo=pre-commit&logoColor=white\n    :target: https://github.com/pre-commit/pre-commit\n\n--------------\n\n**Documentation**: `https://django-swagger-tester.readthedocs.io <https://django-swagger-tester.readthedocs.io/en/latest/?badge=latest>`_\n\n**Repository**: `https://github.com/sondrelg/django-swagger-tester <https://github.com/sondrelg/django-swagger-tester>`_\n\n--------------\n\n\n\nThis package is a simple test utility for your Django Swagger documentation.\n\nIts aim is to make it easy for developers to catch and correct documentation errors in their Swagger docs by\ncomparing documented responses to actual API responses, or validating documented request bodies using actual input serializers.\n\nFeatures\n--------\n\nThe package has three main features:\n\n-  `Testing response documentation`_\n\n-  `Testing input documentation`_\n\n-  `Ensuring your docs comply with a single parameter naming standard`_.\n\n   Supported naming standards include ``camelCase``, ``snake_case``,\n   ``kebab-case``, and ``PascalCase``.\n\n\nImplementations\n---------------\n\nWe currently support testing of:\n\n- Dynamically rendered Swagger docs, using `drf_yasg`_\n- All implementations which render Swagger docs from a schema file (yaml or json)\n\nIf you\'re using another method to generate your documentation and would like to use this package, feel free to add an issue, or create a PR. Adding a new implementation is as easy as adding the required logic needed to load the OpenAPI schema.\n\nInstallation\n============\n\nInstall using pip:\n\n.. code:: python\n\n   pip install django-swagger-tester\n\nConfiguration\n=============\n\nSettings\n--------\n\nTo use Django Swagger Settings in your project, your first need to add a ``SWAGGER_TESTER``\nobject to your ``settings.py``:\n\n.. code:: python\n\n   SWAGGER_TESTER = {\n       \'CASE\': \'camel case\',\n       \'PATH\': BASE_DIR + \'/openapi-schema.yml\'  # not required for drf_yasg implementations\n   }\n\nParameters\n----------\n\n*CASE*\n~~~~~~\n\nThe parameter naming standard you wish to enforce for your documentation.\n\nNeeds to be one of the following:\n\n-  ``camel case``\n-  ``snake case``\n-  ``pascal case``\n-  ``kebab case``\n-  ``None``\n\nYour OpenAPI schema will be assessed to make sure all parameter names\nare correctly cased according to this preference. If you do not wish\nto enforce this check, you can specify ``None`` to skip this feature.\n\nExample:\n\n.. code:: python\n\n  SWAGGER_TESTER = {\n      \'CASE\': \'snake case\',\n  }\n\n**Default**: ``camel case``\n\n*PATH*\n~~~~~~\n\nThe path to your OpenAPI specification.\n\nExample:\n\n.. code:: python\n\n  SWAGGER_TESTER = {\n      \'PATH\': BASE_DIR + \'/openapi-schema.yml\',\n  }\n\n*This setting is not required if your swagger docs are generated.*\n\n*CAMEL_CASE_PARSER*\n~~~~~~~~~~~~~~~~~~~\n\nShould be set to ``True`` if you use `djangorestframework-camel-case <https://github.com/vbabiy/djangorestframework-camel-case>`_\'s\n``CamelCaseJSONParser`` or ``CamelCaseJSONRenderer`` for your API views.\n\nBy settings this to True, example values constructed in the ``validate_input`` function will be snake cased before it\'s passed\nto a serializer. See the `function docs <https://django-swagger-tester.readthedocs.io/en/latest/implementation.html#the-validate-input-function>`_ for more info.\n\nExample:\n\n.. code:: python\n\n  SWAGGER_TESTER = {\n      \'CAMEL_CASE_PARSER\': True,\n  }\n\n\nResponse Validation\n===================\n\nTo verify that your API response documentation is correct, we test the\ngenerated documentation against actual API responses.\n\nA pytest implementation might look like this:\n\n.. code:: python\n\n   from django_swagger_tester.drf_yasg import validate_response  # or replace drf_yasg with `static_schema`\n\n\n   def test_response_documentation(client):\n       response = client.get(\'api/v1/test/\')\n\n       assert response.status_code == 200\n       assert response.json() == expected_response\n\n       # Test Swagger documentation\n       validate_response(response=response, method=\'GET\', route=\'api/v1/test/\')\n\nA Django-test implementation might look like this:\n\n.. code:: python\n\n   class MyApiTest(APITestCase):\n\n       def setUp(self) -> None:\n           user, _ = User.objects.update_or_create(username=\'test_user\')\n           self.client.force_authenticate(user=user)\n           self.path = \'/api/v1/test/\'\n\n       def test_get_200(self) -> None:\n           """\n           Verifies that a 200 is returned for a valid GET request to the /test/ endpoint.\n           """\n           response = self.client.get(self.path, headers={\'Content-Type\': \'application/json\'})\n           expected_response = [...]\n\n           self.assertEqual(response.status_code, 200)\n           self.assertEqual(response.json(), expected_response)\n\n           # Test Swagger documentation\n           validate_response(response=response, method=\'GET\', route=self.path)\n\nYou can also test more than a single response at the time:\n\n.. code:: python\n\n    def test_response_documentation(client):\n        # 201 - Resource created\n        response = client.post(\'api/v1/test/\', data=...)\n        validate_response(response=response, method=\'POST\', route=\'api/v1/test/\')\n\n        # 200 - Idempotency --> if an object exists, return it with a 200 without creating a new resource\n        response = client.post(\'api/v1/test/\', data=...)\n        validate_response(response=response, method=\'POST\', route=\'api/v1/test/\')\n\n        # 400 - Bad data\n        response = client.post(\'api/v1/test/\', data=bad_data)\n        validate_response(response=response, method=\'POST\', route=\'api/v1/test/\')\n\nInput Validation\n================\n\nTo make sure your request body documentation is accurate, and will stay accurate, it can be useful to set up tests.\n\nConsidering most APIs will use input serializers for input validation, it seems sensible to just run the\nexample documentation on your serializer.\n\nA pytest implementation of input validation might look like this::\n\n    from myapp.api.serializers import MyAPISerializer  # <-- your custom serializer\n\n\n    def test_request_body_documentation(client):\n        """\n        Verifies that our request body documentation is representative of a valid request body.\n        """\n        from django_swagger_tester.drf_yasg import validate_input  # or replace drf_yasg with `static_schema`\n        validate_input(serializer=MyAPISerializer, method=\'POST\', route=\'api/v1/test/\', camel_case_parser=True)\n\n\nThe ``camel_case_parser`` argument should be set to ``True`` if you are using ``CamelCaseJSONParser`` or ``CamelCaseJSONRenderer``\nfrom the `djangorestframework-camel-case <https://github.com/vbabiy/djangorestframework-camel-case>`_ package.\n\n.. _`https://django-swagger-tester.readthedocs.io/`: https://django-swagger-tester.readthedocs.io/en/latest/?badge=latest\n.. _Testing response documentation: https://django-swagger-tester.readthedocs.io/en/latest/implementation.html#response-validation\n.. _Testing input documentation: https://django-swagger-tester.readthedocs.io/en/latest/implementation.html#input-validation\n.. _Ensuring your docs comply with a single parameter naming standard: https://django-swagger-tester.readthedocs.io/en/latest/implementation.html#case-checking\n.. _drf_yasg: https://github.com/axnsan12/drf-yasg\n.. _documentation: https://django-swagger-tester.readthedocs.io/\n.. _docs: https://django-swagger-tester.readthedocs.io/\n',
    'author': 'Sondre Lillebø Gundersen',
    'author_email': 'sondrelg@live.no',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/sondrelg/django-swagger-tester',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)



class SwaggerDocumentationError(Exception):
    """
    Custom exception raised when package tests fail.
    """

    pass


class CaseError(Exception):
    """
    Custom exception raised when items are not cased correctly.
    """

    pass


class OpenAPISchemaError(Exception):
    """
    Custom exception raised for invalid schema specifications.
    """

    pass



import json
import logging
import sys
from collections import KeysView
from typing import Any, Optional

from django_swagger_tester.exceptions import SwaggerDocumentationError
from django_swagger_tester.input_validation.utils import serialize_schema
from django_swagger_tester.openapi import index_schema
from django_swagger_tester.utils import replace_refs

logger = logging.getLogger('django_swagger_tester')


def get_response_schema(schema: dict, route: str, method: str, status_code: Optional[int]) -> dict:
    """
    Indexes schema by url, HTTP method, and status code to get the section of a schema related to a specific response.

    :param schema: Full OpenAPI schema
    :param route: Schema-compatible path
    :param method: HTTP request method
    :param status_code: HTTP response code
    :return Response schema
    """
    # Replace all $ref sections in the schema with actual values
    no_ref_schema = replace_refs(schema)
    # Index by paths
    paths_schema = index_schema(schema=no_ref_schema, variable='paths')
    # Index by route
    route_error = f'\n\nFor debugging purposes: valid routes include {", ".join([key for key in paths_schema.keys()])}'
    route_schema = index_schema(schema=paths_schema, variable=route, error_addon=route_error)
    # Index by method
    joined_methods = ', '.join([method.upper() for method in route_schema.keys() if method.upper() != 'PARAMETERS'])
    method_error = f'\n\nAvailable methods include {joined_methods}.'
    method_schema = index_schema(schema=route_schema, variable=method.lower(), error_addon=method_error)
    # Index by responses
    responses_schema = index_schema(schema=method_schema, variable='responses')
    # Index by status code
    status_code_error = f'Documented responses include {", ".join([code for code in responses_schema.keys()])}.'
    status_code_schema = index_schema(schema=responses_schema, variable=str(status_code), error_addon=status_code_error)

    # Not sure about this logic - this is what my static schema looks like, but not the drf_yasg dynamic schema
    if 'content' in status_code_schema and 'application/json' in status_code_schema['content']:
        status_code_schema = status_code_schema['content']['application/json']

    return index_schema(status_code_schema, 'schema')


def format_error(
    error_message: str, data: Any, schema: dict, reference: str, hint: Optional[str] = None, **kwargs
) -> SwaggerDocumentationError:
    """
    Formats and returns a standardized exception and error message.
    """
    logger.debug('Constructing error message')

    if hint is None:
        hint = ''

    # Construct example dict/list from schema - this is useful to display comparable items
    example_item = serialize_schema(schema)

    def get_dotted_line(values: list) -> str:
        longest_value = max(len(f'{v}') for v in values)
        line_length = longest_value if longest_value < 91 else 91
        return line_length * '-' + '\n'

    # Find the max length of keys and values we're showing and create a dotted line to helps us format a nice looking error message
    dotted_line = get_dotted_line(values=[reference, example_item, data])
    longest_key = max(len(i) for i in ['Sequence', 'Expected', 'Received'])

    verbose = 'verbose' in kwargs and kwargs['verbose'] is True
    if verbose:
        tab = (longest_key + 4) * ' '

        # - Construct additional tables to be added onto the message list

        # Unpack schema and data dicts
        schema_items = [{'key': f'{key}', 'value': f'{value}'} for key, value in schema.items()]
        data_items = [{'key': 'data', 'value': f'{data}'}, {'key': 'type', 'value': f'{type(data)}'}]

        # Find length of items
        longest_detailed_key = max(
            len(f'{item["key"]}') for item_list in [schema_items, data_items] for item in item_list
        )

        offset = longest_detailed_key + 4
        addon = ''.join(
            [f'\nResponse details\n', f'{dotted_line}']
            + [item['key'].ljust(offset) + f'{item["value"]}\n' for item in data_items]
            + [f'{dotted_line}\n', f'Schema\n', f'{dotted_line}']
            + [item['key'].ljust(offset) + f'{item["value"]}\n' for item in schema_items]
            + [f'{dotted_line}']
        )

        # Then - For a detailed view, we change `example item` and `data` to expanded versions of themselves
        example_item: str = f'\n{tab}' + json.dumps(example_item, indent=4, sort_keys=True).replace('\n', f'\n{tab}')  # type: ignore
        data: str = f'\n{tab}' + json.dumps(data, indent=4, sort_keys=True).replace('\n', f'\n{tab}')  # type: ignore

    else:
        addon = '\n* If you need more details: set `verbose=True`'

    sys.stdout.flush()

    # Construct error message
    offset = longest_key + 4
    message = [
        f'Item is misspecified:\n\n'
        # -- Summary table --
        f'Summary',
        '\n' if not verbose else '',
        f'{dotted_line}',
        '\n',
        'Error:'.ljust(offset) + f'{error_message}\n',
        '\n',
        'Expected:'.ljust(offset) + f'{example_item}\n',
        'Received:'.ljust(offset) + f'{data}\n',
        '\n',
        'Hint:'.ljust(offset)
        + '\n'.ljust(offset + 1).join(hint.split('\n'))
        + '\n',  # the join logic adds support for multi-line hints
        'Sequence:'.ljust(offset) + f'{reference}\n',
        '\n' if not verbose else '',
        f'{dotted_line}',
        f'{addon}',
    ]

    return SwaggerDocumentationError(''.join(message))


def check_keys_match(schema_keys: KeysView, response_keys: KeysView, schema: dict, data: dict, reference: str) -> None:
    """
    Verifies that both sets have the same amount of keys.
    A length mismatch in the two sets, indicates an error in one of them.

    :param schema_keys: Schema object keys
    :param response_keys: Response dictionary keys
    :param schema: OpenAPI schema
    :param data: Response data
    :param reference: Logging reference to output for errors -
        this makes it easier to identify where in a response/schema an error is happening
    :raises: django_swagger_tester.exceptions.SwaggerDocumentationError
    """
    if len(schema_keys) != len(response_keys):
        logger.debug('The number of schema dict elements does not match the number of response dict elements')
        if len(set(response_keys)) > len(set(schema_keys)):
            missing_keys = ', '.join([f'`{key}`' for key in list(set(response_keys) - set(schema_keys))])
            raise format_error(
                f'The following properties seem to be missing from your OpenAPI/Swagger documentation: {missing_keys}.',
                data=data,
                schema=schema,
                reference=reference,
                hint='Add the key(s) to your Swagger docs, or stop returning it in your view.',
            )
        else:
            missing_keys = ', '.join([f'{key}' for key in list(set(schema_keys) - set(response_keys))])
            raise format_error(
                f'The following properties seem to be missing from your response body: {missing_keys}.',
                data=data,
                schema=schema,
                reference=reference,
                hint='Remove the key(s) from you Swagger docs, or include it in your API response.',
            )



import logging
from json import dumps, loads
from typing import Optional

from django.apps import apps
from django.core.exceptions import ImproperlyConfigured

from django_swagger_tester.input_validation.utils import get_request_body
from django_swagger_tester.response_validation.utils import get_response_schema
from django_swagger_tester.utils import get_paths, resolve_path, validate_inputs

logger = logging.getLogger('django_swagger_tester')


# noinspection PyMethodMayBeStatic
class LoadDrfYasgSchema:
    def __init__(self, route: str, method: str, status_code: Optional[int] = None, **kwargs) -> None:
        """
        Loads OpenAPI schema when schema is dynamically generated by drf_yasg.

        :param route: a django-resolved endpoint path
        :param status_code: the relevant HTTP response status code to check in the OpenAPI schema
        :param method: the relevant HTTP method to check in the OpenAPI schema
        """
        validate_inputs(route=route, status_code=status_code, method=method)
        self.validation()

        from drf_yasg.openapi import Info
        from drf_yasg.generators import OpenAPISchemaGenerator

        self.schema_generator = OpenAPISchemaGenerator(info=Info(title='', default_version=''))

        self.route = self.get_drf_yasg_compatible_route(route)
        self.status_code = status_code
        self.method = method

    def validation(self) -> None:
        """
        For drf_yasg-generated schemas, it's important that we verify:
        1. The package is installed
        2. Json is installed, for parsing the schema
        3. drf_yasg is in the projects installed_apps
        """
        try:
            import drf_yasg  # noqa: F401
        except ModuleNotFoundError:
            raise ImproperlyConfigured(
                'The package `drf_yasg` is required. Please run `pip install drf_yasg` to install it.'
            )

        if 'drf_yasg' not in apps.app_configs.keys():
            raise ImproperlyConfigured(
                'The package `drf_yasg` is missing from INSTALLED_APPS. Please add it to your '
                '`settings.py`, as it is required for this implementation'
            )

    def get_schema(self) -> dict:
        """
        Generates OpenAPI schema.
        """
        odict_schema = self.schema_generator.get_schema(None, True)
        return loads(dumps(odict_schema.as_odict()))

    def get_path_prefix(self) -> str:
        """
        Returns the drf_yasg specified path prefix.

        Drf_yasg `cleans` schema paths by finding recurring path patterns,
        and cutting them out of the generated openapi schema.
        For example, `/api/v1/example` might then just become `/example`
        """
        return self.schema_generator.determine_path_prefix(get_paths())

    def get_drf_yasg_compatible_route(self, route: str) -> str:
        """
        Returns a url that matches the urls found in a drf_yasg-generated schema.

        :param route: Django resolved route
        """
        resolved_route = resolve_path(route)
        path_prefix = self.get_path_prefix()  # typically might be 'api/' or 'api/v1/'
        logger.debug('Path prefix: %s', path_prefix)
        if path_prefix != '/':
            return resolved_route[len(path_prefix) :]
        else:
            return resolved_route

    def get_response_schema(self) -> dict:
        """
        Indexes schema by url, HTTP method, and status code
        to get the section of a schema related to a specific response.
        """
        schema = self.get_schema()
        return get_response_schema(schema=schema, method=self.method, status_code=self.status_code, route=self.route)

    def get_request_body(self) -> dict:
        """
        Indexes schema by url, HTTP method, and status code,
        to get the request body of the section of a schema related to an endpoint.
        """
        schema = self.get_schema()
        return get_request_body(schema=schema, method=self.method, route=self.route)
